@@ -1,19 +1,7 @@
+// SPDX-License-Identifier: GPL-2.0
 /*
  * Copyright (c) 2000-2005 Silicon Graphics, Inc.
  * All Rights Reserved.
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation.
- *
- * This program is distributed in the hope that it would be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write the Free Software Foundation,
- * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #include "libxfs.h"
@@ -236,11 +224,11 @@ fp_dinode_fmt(
 	int			array)
 {
 	int			bitpos;
-	xfs_dinode_fmt_t	f;
+	enum xfs_dinode_fmt	f;
 	int			i;
 
 	for (i = 0, bitpos = bit; i < count; i++, bitpos += size) {
-		f = (xfs_dinode_fmt_t)getbitval(obj, bitpos, size, BVUNSIGNED);
+		f = (enum xfs_dinode_fmt)getbitval(obj, bitpos, size, BVUNSIGNED);
 		if (array)
 			dbprintf("%d:", i + base);
 		if (f < 0 || f >= dinode_fmt_name_size)
@@ -650,6 +638,7 @@ set_cur_inode(
 	int		offset;
 	int		numblks = blkbb;
 	xfs_agblock_t	cluster_agbno;
+	struct xfs_ino_geometry *igeo = M_IGEO(mp);
 
 
 	agno = XFS_INO_TO_AGNO(mp, ino);
@@ -664,20 +653,18 @@ set_cur_inode(
 	}
 	cur_agno = agno;
 
-	if (mp->m_inode_cluster_size > mp->m_sb.sb_blocksize &&
-	    mp->m_inoalign_mask) {
+	if (igeo->inode_cluster_size > mp->m_sb.sb_blocksize &&
+	    igeo->inoalign_mask) {
 		xfs_agblock_t	chunk_agbno;
 		xfs_agblock_t	offset_agbno;
-		int		blks_per_cluster;
 
-		blks_per_cluster = mp->m_inode_cluster_size >>
-							mp->m_sb.sb_blocklog;
-		offset_agbno = agbno & mp->m_inoalign_mask;
+		offset_agbno = agbno & igeo->inoalign_mask;
 		chunk_agbno = agbno - offset_agbno;
 		cluster_agbno = chunk_agbno +
-			((offset_agbno / blks_per_cluster) * blks_per_cluster);
+			((offset_agbno / M_IGEO(mp)->blocks_per_cluster) *
+			 M_IGEO(mp)->blocks_per_cluster);
 		offset += ((agbno - cluster_agbno) * mp->m_sb.sb_inopblock);
-		numblks = XFS_FSB_TO_BB(mp, blks_per_cluster);
+		numblks = XFS_FSB_TO_BB(mp, M_IGEO(mp)->blocks_per_cluster);
 	} else
 		cluster_agbno = agbno;
 
