@@ -46,14 +46,14 @@
 #include "cache.h"
 #include "bitops.h"
 #include "kmem.h"
-#include "radix-tree.h"
+#include "libfrog/radix-tree.h"
 #include "atomic.h"
 
 #include "xfs_types.h"
 #include "xfs_arch.h"
 
 #include "xfs_fs.h"
-#include "crc32c.h"
+#include "libfrog/crc32c.h"
 
 /* Zones used in libxfs allocations that aren't in shared header files */
 extern kmem_zone_t *xfs_buf_item_zone;
@@ -62,9 +62,18 @@ extern kmem_zone_t *xfs_buf_zone;
 extern kmem_zone_t *xfs_inode_zone;
 extern kmem_zone_t *xfs_trans_zone;
 
+#define timespec64 timespec
+
+/* fake up iomap, (not) used in xfs_bmap.[ch] */
+#define IOMAP_F_SHARED			0x04
+#define xfs_bmbt_to_iomap(a, b, c, d)	((void) 0)
+
 /* CRC stuff, buffer API dependent on it */
 #define crc32c(c,p,l)	crc32c_le((c),(unsigned char const *)(p),(l))
 
+/* fake up kernel's iomap, (not) used in xfs_bmap.[ch] */
+struct iomap;
+
 #include "xfs_cksum.h"
 
 /*
@@ -139,6 +148,14 @@ enum ce { CE_DEBUG, CE_CONT, CE_NOTE, CE_WARN, CE_ALERT, CE_PANIC };
 	cmn_err(CE_ALERT, "%s: XFS_ERROR_REPORT", (e));  \
 } while (0)
 
+#define XFS_WARN_CORRUPT(mp, expr) \
+	( ((mp)->m_flags & LIBXFS_MOUNT_WANT_CORRUPTED) ? \
+	   (printf("%s: XFS_WARN_CORRUPT at %s:%d", #expr, \
+		   __func__, __LINE__), true) : true)
+
+#define XFS_IS_CORRUPT(mp, expr)	\
+	(unlikely(expr) ? XFS_WARN_CORRUPT((mp), (expr)) : false)
+
 #define XFS_ERRLEVEL_LOW		1
 #define XFS_FORCED_SHUTDOWN(mp)		0
 #define XFS_ILOCK_EXCL			0
@@ -146,25 +163,6 @@ enum ce { CE_DEBUG, CE_CONT, CE_NOTE, CE_WARN, CE_ALERT, CE_PANIC };
 #define XFS_STATS_DEC(mp, count, x)	do { (mp) = (mp); } while (0)
 #define XFS_STATS_ADD(mp, count, x)	do { (mp) = (mp); } while (0)
 #define XFS_TEST_ERROR(expr,a,b)	( expr )
-#define XFS_WANT_CORRUPTED_GOTO(mp, expr, l)				\
-{									\
-	if (!(expr)) {							\
-		if ((mp)->m_flags & LIBXFS_MOUNT_WANT_CORRUPTED)	\
-			printf("WANT_CORRUPTED_GOTO at %s:%d\n",	\
-				__func__, __LINE__);			\
-		error = -EFSCORRUPTED;					\
-		goto l;							\
-	}								\
-}
-#define XFS_WANT_CORRUPTED_RETURN(mp, expr)				\
-{									\
-	if (!(expr)) {							\
-		if ((mp)->m_flags & LIBXFS_MOUNT_WANT_CORRUPTED)	\
-			printf("WANT_CORRUPTED_RETURN at %s:%d\n",	\
-				__func__, __LINE__);			\
-		return -EFSCORRUPTED;					\
-	}								\
-}
 
 #ifdef __GNUC__
 #define __return_address	__builtin_return_address(0)
@@ -379,13 +377,6 @@ roundup_64(uint64_t x, uint32_t y)
 	(len) = __bar; /* no set-but-unused warning */	\
 	NULL;						\
 })
-#define xfs_buf_get_uncached(t,n,f)     \
-	libxfs_getbufr((t), XFS_BUF_DADDR_NULL, (n));
-#define xfs_buf_relse(bp)		libxfs_putbuf(bp)
-#define xfs_buf_get(devp,blkno,len,f)	(libxfs_getbuf((devp), (blkno), (len)))
-#define xfs_bwrite(bp)			libxfs_writebuf((bp), 0)
-#define xfs_buf_delwri_queue(bp, bl)	libxfs_writebuf((bp), 0)
-#define xfs_buf_delwri_submit(bl)	(0)
 #define xfs_buf_oneshot(bp)		((void) 0)
 
 #define XBRW_READ			LIBXFS_BREAD
@@ -536,6 +527,8 @@ void xfs_inode_verifier_error(struct xfs_inode *ip, int error,
 
 #define xfs_buf_verifier_error(bp,e,n,bu,bus,fa) \
 	xfs_verifier_error(bp, e, fa)
+void
+xfs_buf_corruption_error(struct xfs_buf *bp);
 
 /* XXX: this is clearly a bug - a shared header needs to export this */
 /* xfs_rtalloc.c */
@@ -611,14 +604,6 @@ static inline int test_and_set_bit(int nr, volatile unsigned long *addr)
 }
 
 /* Keep static checkers quiet about nonstatic functions by exporting */
-int xfs_inode_hasattr(struct xfs_inode *ip);
-int xfs_attr_get_ilocked(struct xfs_inode *ip, struct xfs_da_args *args);
-int xfs_attr_get(struct xfs_inode *ip, const unsigned char *name,
-                unsigned char *value, int *valuelenp, int flags);
-int xfs_attr_set(struct xfs_inode *dp, const unsigned char *name,
-                 unsigned char *value, int valuelen, int flags);
-int xfs_attr_remove(struct xfs_inode *dp, const unsigned char *name, int flags);
-
 int xfs_rtbuf_get(struct xfs_mount *mp, struct xfs_trans *tp,
 		  xfs_rtblock_t block, int issum, struct xfs_buf **bpp);
 int xfs_rtcheck_range(struct xfs_mount *mp, struct xfs_trans *tp,
